#ifndef INCLUDED_MOON_INPUT_H
#define INCLUDED_MOON_INPUT_H

#include <SDL.h>

#define SWIDTH  320
#define SHEIGHT 240
#define BORDER  5
#define BWIDTH  (SWIDTH+2*BORDER)
#define BHEIGHT (SHEIGHT+2*BORDER)
#define TWIDTH  32
#define THEIGHT 32
#define TCX     (TWIDTH/2)
#define TCY     (THEIGHT/2)

#define LOGICMS 50
#define LOGICTIME (1000/50)
#define PAUSETIME 2500

#define UNUSED __attribute__((__unused__))

#define ANIMFXPBITS 4
#define ANIMSPEED   2
#define ANIMFRAMES  2

#ifdef GP2X
#define GP2X_BUTTON_UP              (0)
#define GP2X_BUTTON_DOWN            (4)
#define GP2X_BUTTON_LEFT            (2)
#define GP2X_BUTTON_RIGHT           (6)
#define GP2X_BUTTON_UPLEFT          (1)
#define GP2X_BUTTON_UPRIGHT         (7)
#define GP2X_BUTTON_DOWNLEFT        (3)
#define GP2X_BUTTON_DOWNRIGHT       (5)
#define GP2X_BUTTON_CLICK           (18)
#define GP2X_BUTTON_A               (12)
#define GP2X_BUTTON_B               (13)
#define GP2X_BUTTON_X               (14)
#define GP2X_BUTTON_Y               (15)
#define GP2X_BUTTON_L               (10)
#define GP2X_BUTTON_R               (11)
#define GP2X_BUTTON_START           (8)
#define GP2X_BUTTON_SELECT          (9)
#define GP2X_BUTTON_VOLUP           (16)
#define GP2X_BUTTON_VOLDOWN         (17)
#endif

typedef struct {
	unsigned int n;
	unsigned int ne;
	unsigned int e;
	unsigned int se;
	unsigned int s;
	unsigned int sw;
	unsigned int w;
	unsigned int nw;
	unsigned int start;
	unsigned int select;
	unsigned int fire1;
	unsigned int fire2;
	unsigned int cancel;
} buttons_t;

typedef unsigned int (*logicfn_t) (unsigned int frameno);
typedef void (*videofn_t) (unsigned int delta);

extern SDL_Surface *surf;
extern buttons_t buttons;
extern logicfn_t logicfn;
extern videofn_t videofn;

typedef enum {
	Einitsdl = 1,
	Einitpfs,
	Econfpfs,
	Evideomode,
	Esurface,
	Eload,
	Eoom
} error_t;

void error (error_t e);

void go_title (void);
void go_game (unsigned int room);
SDL_Surface *load (const char *file, unsigned int level);
void loadgamegfx (void);
void unloadgamegfx (void);

static inline void freesurf (SDL_Surface **s)
{
	if (*s)
	{
		SDL_FreeSurface(*s);
		*s = NULL;
	}
}

static inline unsigned int uabs (int a)
{
	return a < 0 ? (unsigned int)-a : (unsigned int)a;
}

static inline int imin (int a, int b)
{
	return a < b ? a : b;
}

static inline int imax (int a, int b)
{
	return a >= b ? a : b;
}

#endif
