#include "conf.h"

typedef enum {
	EntRing = 1,
	EntGate_NS,
	EntGate_EW,
	EntDiver_N,
	EntDiver_S,
	EntShooter,
	ENTCOUNT,
	EntMu_E = ENTCOUNT,
	EntMu_W,
	EntShot,
	ENTCOUNT2
} etype_t;

static const char *entnames[] = {
	"ring---",
	"gate-ns",
	"gate-ew",
	"diver-n",
	"diver-s",
	"shooter",
	"mu----e",
	"mu----w",
	"shot---"
};

typedef struct ent_t ent_t;
struct ent_t {
	ent_t *next;
	etype_t type;
	unsigned int foff;
	int x;
	int y;
	unsigned int state;
};

static struct {
	SDL_Surface *walls, *back, *mumask, *shotm;
	SDL_Surface *ent[ENTCOUNT2][ANIMFRAMES];
} gfx = { NULL, NULL, NULL, NULL };

static Uint8 state[0x10000];
#define GotRings 1

static ent_t *ent = NULL, *player = NULL;

static unsigned int current = 0x8080, aframe = 0;
static int lastx = 30, lasty = 30;
static etype_t lastt = EntMu_E;

static unsigned int rings = 0;

static inline char hexit (unsigned int num)
{
	return num < 10 ? num + '0' : num - 10 + 'a';
}

static void clearents (void)
{
	while (ent != player)
	{
		ent_t *tmp = ent;
		ent = ent->next;
		free(tmp);
	}
}

static void opengates (void)
{
	ent_t *eit = ent, *prev = NULL;
	while (eit != player)
	{
		if (eit->type == EntGate_NS || eit->type == EntGate_EW)
		{
			ent_t *tmp = eit;
			if (prev)
				prev->next = eit->next;
			else
				ent = eit->next;
			eit = eit->next;
			free(tmp);
		}
		else
		{
			prev = eit;
			eit = eit->next;
		}
	}
}

static void byeshot (ent_t *shot)
{
	ent_t *eit = ent, *prev = NULL;
	while (eit != shot)
	{
		prev = eit;
		eit = eit->next;
	}
	if (prev)
		prev->next = eit->next;
	else
		ent = eit->next;
	free(shot);
}

static void addents (SDL_Surface *entmap)
{
	const unsigned int mpt = entmap->pitch;
	const Uint8 *map = &(((Uint8*)entmap->pixels)[BHEIGHT * mpt]);
	ent_t *newent;
	unsigned int y = BHEIGHT;

	if (!player)
	{
		if (!(newent = (ent_t*)malloc(sizeof(ent_t)))) error(Eoom);
		newent->type = lastt;
		newent->foff = 0;
		newent->x = lastx;
		newent->y = lasty;
		newent->state = 0;
		newent->next = NULL;
		player = ent = newent;
	}

	rings = 0;

	while (y--)
	{
		unsigned int x = BWIDTH;
		map -= mpt;
		while (x--)
		{
			if (map[x])
			{
				ent_t *newent;
				if (map[x] == EntRing && (state[current] & GotRings))
					continue;
				newent = (ent_t*)malloc(sizeof(ent_t));
				if (!newent) error(Eoom);
				newent->type = map[x];
				newent->foff = (map[x] == EntGate_NS || map[x] == EntGate_EW)
					? 0 : (unsigned int)(rand() / (RAND_MAX/10));
				newent->x = (int)x - TCX;
				newent->y = (int)y - TCY;
				newent->state = 0;
				newent->next = ent;
				ent = newent;
				if (newent->type == EntRing)
					++rings;
			}
		}
	}
	if (!rings)
		opengates();
	SDL_FreeSurface(entmap);
}

static void spawnshot (ent_t *parent)
{
	ent_t *newent = (ent_t*)malloc(sizeof(ent_t));
	if (!newent) error(Eoom);
	newent->type = EntShot;
	newent->foff = (unsigned int)(rand() / (RAND_MAX/10));
	newent->x = parent->x;
	newent->y = parent->y;
	newent->state = parent->state;
	newent->next = ent;
	ent = newent;
}

static void go_room (unsigned int room, int reset)
{
	char wimg[] = "levels/0000w.png";
	char *wroom = wimg + 7;
	char *wtype = wimg + 11;
	freesurf(&gfx.walls);
	freesurf(&gfx.back);
	if (ent) clearents();
	if (room == 0x807f)
	{
		go_title();
		return;
	}
	if (reset)
	{
		if (player)
		{
			free(player);
			player = NULL;
		}
	}
	else
	{
		lastx = player->x;
		lasty = player->y;
		lastt = player->type;
	}
	current = room;
	wroom[3] = hexit(room & 0x0f);
	room >>= 4;
	wroom[2] = hexit(room & 0x0f);
	room >>= 4;
	wroom[1] = hexit(room & 0x0f);
	room >>= 4;
	wroom[0] = hexit(room & 0x0f);
	gfx.walls = load(wimg, 1);
	*wtype = 'b';
	gfx.back = load(wimg, 1);
	*wtype = 'e';
	addents(load(wimg, 1));
}

static inline unsigned int overlap (ent_t *e1, ent_t *e2)
{
	const int x1 = e1->x, y1 = e1->y, x2 = e2->x, y2 = e2->y;
	return !(x1 + TWIDTH < x2 || x1 >= x2 + TWIDTH || y1 + THEIGHT < y2 || y1 >= y2 + THEIGHT);
}

static SDL_Surface *getentimage (ent_t *e)
{
	unsigned int eframe = (e->foff + aframe) >> ANIMFXPBITS;
	if (eframe >= ANIMFRAMES) eframe -= ANIMFRAMES;
	return gfx.ent[e->type][eframe];
}

static unsigned int collideent (ent_t *e1, ent_t *e2)
{
	SDL_Surface *e1s = getentimage(e1), *e2s = getentimage(e2);
	const unsigned int e1pt = e1s->pitch, e2pt = e2s->pitch;

	int x0 = imax(e1->x, e2->x), y0 = imax(e1->y, e2->y);
	const int ow = imin(e1->x, e2->x) + TWIDTH - x0, oh = imin(e1->y, e2->y) + THEIGHT - y0;

	Uint8 *e1px = &(((Uint8*)e1s->pixels)[(y0 - e1->y) * e1pt + x0 - e1->x]);
	Uint8 *e2px = &(((Uint8*)e2s->pixels)[(y0 - e2->y) * e2pt + x0 - e2->x]);

	int y = oh;

	while (y--)
	{
		int x = ow;
		while (x--)
			if (e1px[x] && e2px[x])
				return 1;
		e1px += e1pt;
		e2px += e2pt;
	}

	return 0;
}

static unsigned int collidebg (SDL_Surface *tmask, int x0, int y0)
{
	const int x1 = x0 + TWIDTH, y1 = y0 + THEIGHT;
	int xs = 0, ys = 0, xe = TWIDTH, ye = THEIGHT;
	if (x0 < 0)
	{
		xs -= x0;
		x0 = 0;
	}
	else if (x1 >= BWIDTH)
		xe -= x1 - BWIDTH;
	if (y0 < 0)
	{
		ys -= y0;
		y0 = 0;
	}
	else if (y1 >= BHEIGHT)
		ye -= y1 - BHEIGHT;

	{
	const unsigned int wpt = gfx.walls->pitch, mpt = tmask->pitch;
	Uint8 *wpx = &(((Uint8*)gfx.walls->pixels)[y0 * wpt + x0]);
	Uint8 *mpx = &(((Uint8*)tmask->pixels)[ys * mpt]);
	int x, y;

	for (y = ys; y < ye; ++y)
	{
		for (x = xs; x < xe; ++x)
		{
			if (wpx[x] & mpx[x])
				return 1;
		}
		wpx += wpt;
		mpx += mpt;
	}
	}

	return 0;
}

static void entlogic (ent_t *ent, unsigned int frameno)
{
	SDL_Surface *mask = NULL;
	unsigned int cbg = 0;
	int x = ent->x, y = ent->y;

	switch (ent->type)
	{
		case EntDiver_N:
			mask = gfx.mumask;
			if (!--y)
				ent->type = EntDiver_S;
			break;
		case EntDiver_S:
			mask = gfx.mumask;
			if (++y >= BHEIGHT)
				ent->type = EntDiver_N;
			break;
		case EntShooter:
			if (!(frameno & 0x0f))
			{
				if (++ent->state >= 8)
					ent->state = 0;
				spawnshot(ent);
			}
			break;
		case EntShot:
			mask = gfx.shotm;
			switch (ent->state)
			{
				case 1:
					++x;
				case 0:
					--y;
					break;
				case 3:
					++y;
				case 2:
					++x;
					break;
				case 5:
					--x;
				case 4:
					++y;
					break;
				case 7:
					--y;
				case 6:
					--x;
					break;
				default: break;
			}
		default: break;
	}

	if (mask)
		cbg = collidebg(mask, x, y);

	if (cbg & 1) switch (ent->type)
	{
		case EntDiver_N:
			ent->type = EntDiver_S;
			break;
		case EntDiver_S:
			ent->type = EntDiver_N;
			break;
		case EntShot:
			byeshot(ent);
			break;
		default: break;
	}
	else
	{
		ent->x = x;
		ent->y = y;
	}
}

static unsigned int logicframe (unsigned int frameno)
{
	unsigned int newroom = current;
	int nplx = player->x, nply = player->y;

	if ((aframe += ANIMSPEED) >= (ANIMFRAMES<<ANIMFXPBITS))
		aframe -= (ANIMFRAMES<<ANIMFXPBITS);

	if (buttons.n) --nply;
	if (buttons.s) ++nply;
	if (collidebg(gfx.mumask, nplx, nply))
		nply = player->y;

	if (buttons.e) ++nplx;
	if (buttons.w) --nplx;
	if (collidebg(gfx.mumask, nplx, nply))
		nplx = player->x;

	if (nplx - player->x > 0)
		player->type = EntMu_E;
	else if (nplx - player->x < 0)
		player->type = EntMu_W;

	if (nplx < -TCX+BORDER)
	{
		newroom -= 0x0100;
		nplx += SWIDTH;
	}
	else if (nplx > BWIDTH-TCX-BORDER)
	{
		newroom += 0x0100;
		nplx -= SWIDTH;
	}
	if (nply < -TCY+BORDER)
	{
		newroom -= 0x0001;
		nply += SHEIGHT;
	}
	else if (nply > BHEIGHT-TCY-BORDER)
	{
		newroom += 0x0001;
		nply -= SHEIGHT;
	}

	player->x = nplx;
	player->y = nply;

	if (newroom == current)
	{
		ent_t *eit, *prev = NULL;
		for (eit = ent; eit;)
		{
			ent_t *tmp = eit->next;
			entlogic(eit, frameno);
			eit = tmp;
		}
		for (eit = ent; eit != player; eit = eit->next)
		{
			if (overlap(eit, player) && collideent(eit, player))
			{
				if (eit->type == EntRing)
				{
					if (prev)
						prev->next = eit->next;
					else
						ent = eit->next;
					free(eit);
					if (!--rings)
					{
						state[current] |= GotRings;
						opengates();
					}
				}
				else go_room(current, 1);
				break;
			}
			prev = eit;
		}
	}
	else go_room(newroom, 0);

	return buttons.cancel ? 0 : 1;
}

static void videoframe (unsigned int delta UNUSED)
{
	ent_t *eit;
	SDL_Rect pos;
	SDL_BlitSurface(gfx.back, NULL, surf, NULL);
	for (eit = ent; eit; eit = eit->next)
	{
		pos.x = eit->x;
		pos.y = eit->y;
		SDL_BlitSurface(getentimage(eit), NULL, surf, &pos);
	}
}

void loadgamegfx (void)
{
	char efile[] = "gfx/--------.png";
	char *ename = efile + 4;
	char *frame = ename + 7;
	unsigned int i = ENTCOUNT2;
	gfx.mumask = load("gfx/mumove.png", 0);
	gfx.shotm = load("gfx/shotmove.png", 0);
	while (--i)
	{
		unsigned int f;
		memcpy(ename, entnames[i-1], 7);
		*frame = '0';
		for (f = 0; f < ANIMFRAMES; ++f)
		{
			gfx.ent[i][f] = load(efile, 0);
			++*frame;
		}
	}
}

void unloadgamegfx (void)
{
	unsigned int i = ENTCOUNT2;
	while (i--)
	{
		unsigned int j = ANIMFRAMES;
		while (j) freesurf(&gfx.ent[i][--j]);
	}
	freesurf(&gfx.walls);
	freesurf(&gfx.mumask);
}

void go_game (unsigned int room)
{
	aframe = 0;
	memset(state, 0, sizeof(state));
	go_room(room, 1);

	logicfn = logicframe;
	videofn = videoframe;
}
